create or replace procedure create_views_kladr(in_version in varchar2) is
begin
  execute immediate 'create materialized view mv'|| in_version ||'_kladr_parsed
    logging nocompress nocache noparallel build immediate never refresh
    as 
    select
        kl.s_code,
        kl.s_identcode, kl.s_identparent,
        kl.s_codestate, (case when kl.s_codestate = ''00'' then 1 else 0 end) as is_actual,
        kl.s_actualcode,
        kl.s_name,
        kl.n_level,
        kl.s_shorttype, sk.c_socrname as s_fulltype,
        kl.s_postindex,
        kl.s_gninmb, kl.s_uno, kl.s_okato, kl.s_status
    from (
      select
        kl.*,
        (case
            when s_codestate = ''00'' then kl.s_code
            when s_codestate >= ''01'' and s_codestate <= ''50'' then kl.s_identcode||''00''
            when s_codestate = ''51'' then (select s_newcodes from (select substr(c_oldcode, 1, 11) as s_oldcode, listagg(c_newcode, '', '') within group (order by c_newcode) as s_newcodes from (select ch.c_oldcode, ch.c_newcode from v'|| in_version ||'_changes ch where length(ch.c_oldcode) = 13) group by c_oldcode) where s_oldcode = kl.s_identcode)
            when s_codestate >= ''52'' and s_codestate <= ''98'' then null
            when s_codestate = ''99'' then null
            else null
        end) as s_actualcode
      from (
        select
            kl.c_code as s_code,
            substr(kl.c_code, 1, 11) as s_identcode,
            (case
                when substr(kl.c_code, 9, 3) > 0 then substr(kl.c_code, 1, 8)||''000''
                when substr(kl.c_code, 6, 3) > 0 then substr(kl.c_code, 1, 5)||''000000''
                when substr(kl.c_code, 3, 3) > 0 then substr(kl.c_code, 1, 2)||''000000000''
                else null
            end) as s_identparent,
            substr(kl.c_code, -2) as s_codestate,
            (case
                when substr(kl.c_code, 9, 3) > 0 then 4
                when substr(kl.c_code, 6, 3) > 0 then 3
                when substr(kl.c_code, 3, 3) > 0 then 2
                else 1
            end) as n_level,
            kl.c_name as s_name,
            kl.c_socr as s_shorttype,
            kl.c_index as s_postindex,
            kl.c_gninmb as s_gninmb,
            kl.c_uno as s_uno,
            kl.c_ocatd as s_okato,
            kl.c_status as s_status
        from v'|| in_version ||'_kladr kl
      ) kl
    ) kl, v'|| in_version ||'_sokr sk
    where kl.s_shorttype = sk.c_scname and kl.n_level = sk.c_level';


  execute immediate 'create materialized view mv'|| in_version ||'_kladr
    logging nocompress nocache noparallel build immediate never refresh
    as
    select
        to_number(kp.s_identcode) as id,
        kp.s_code, kp.s_identcode, kp.s_identparent,
        kp.s_codestate, kp.is_actual, kp.s_actualcode,
        kp.s_name,
        kp.n_level,
        kp.s_shorttype, kp.s_fulltype,
        kp.s_postindex,
        kp.s_gninmb, kp.s_uno, kp.s_okato, kp.s_status
    from (select kp.*, row_number() over(partition by s_identcode order by decode(kp.s_codestate, ''00'', ''100'', kp.s_codestate) desc) as n_order_act from mv'|| in_version ||'_kladr_parsed kp) kp
    where kp.n_order_act = 1';


  execute immediate 'create materialized view mv'|| in_version ||'_kladr_inact
    logging nocompress nocache noparallel build immediate never refresh
    as
    select
        to_number(kp.s_code) as id,
        kp.s_code, kp.s_identcode, kp.s_identparent,
        kp.s_codestate, kp.s_actualcode,
        kp.s_name,
        kp.n_level,
        kp.s_shorttype, kp.s_fulltype,
        kp.s_postindex,
        kp.s_gninmb, kp.s_uno, kp.s_okato, kp.s_status
    from mv'|| in_version ||'_kladr_parsed kp
    where kp.is_actual = 0';
end;



create or replace procedure create_views_street(in_version in varchar2) is
begin
  execute immediate 'create materialized view mv'|| in_version ||'_street_parsed
    logging nocompress nocache noparallel build immediate never refresh
    as 
    select
        kl.s_code,
        kl.s_identcode, kl.s_identparent,
        kl.s_codestate, (case when kl.s_codestate = ''00'' then 1 else 0 end) as is_actual,
        kl.s_actualcode,
        kl.s_name,
        kl.n_level,
        kl.s_shorttype, sk.c_socrname as s_fulltype,
        kl.s_postindex,
        kl.s_gninmb, kl.s_uno, kl.s_okato
    from (
      select
        kl.*,
        (case
            when s_codestate = ''00'' then kl.s_code
            when s_codestate >= ''01'' and s_codestate <= ''50'' then kl.s_identcode||''00''
            when s_codestate = ''51'' then (select s_newcodes from (select substr(c_oldcode, 1, 15) as s_oldcode, listagg(c_newcode, '', '') within group (order by c_newcode) as s_newcodes from (select ch.c_oldcode, ch.c_newcode from v'|| in_version ||'_changes ch where length(ch.c_oldcode) = 17) group by c_oldcode) where s_oldcode = kl.s_identcode)
            when s_codestate >= ''52'' and s_codestate <= ''98'' then null
            when s_codestate = ''99'' then null
            else null
        end) as s_actualcode
      from (
        select
            kl.c_code as s_code,
            substr(kl.c_code, 1, 15) as s_identcode,
            substr(kl.c_code, 1, 11) as s_identparent,
            substr(kl.c_code, -2) as s_codestate,
            5 as n_level,
            kl.c_name as s_name,
            kl.c_socr as s_shorttype,
            kl.c_index as s_postindex,
            kl.c_gninmb as s_gninmb,
            kl.c_uno as s_uno,
            kl.c_ocatd as s_okato
        from v'|| in_version ||'_street kl
      ) kl
    ) kl, v'|| in_version ||'_sokr sk
    where kl.s_shorttype = sk.c_scname and kl.n_level = sk.c_level';


  execute immediate 'create materialized view mv'|| in_version ||'_street
    logging nocompress nocache noparallel build immediate never refresh
    as
    select
        to_number(kp.s_identcode) as id,
        kp.s_code, kp.s_identcode, kp.s_identparent,
        kp.s_codestate, kp.is_actual, kp.s_actualcode,
        kp.s_name,
        kp.n_level,
        kp.s_shorttype, kp.s_fulltype,
        kp.s_postindex,
        kp.s_gninmb, kp.s_uno, kp.s_okato
    from (select kp.*, row_number() over(partition by s_identcode order by decode(kp.s_codestate, ''00'', ''100'', kp.s_codestate) desc) as n_order_act from mv'|| in_version ||'_street_parsed kp) kp
    where kp.n_order_act = 1';


  execute immediate 'create materialized view mv'|| in_version ||'_street_inact
    logging nocompress nocache noparallel build immediate never refresh
    as
    select
        to_number(kp.s_code) as id,
        kp.s_code, kp.s_identcode, kp.s_identparent,
        kp.s_codestate, kp.s_actualcode,
        kp.s_name,
        kp.n_level,
        kp.s_shorttype, kp.s_fulltype,
        kp.s_postindex,
        kp.s_gninmb, kp.s_uno, kp.s_okato
    from mv'|| in_version ||'_street_parsed kp
    where kp.is_actual = 0';
end;


create or replace procedure get_diff(
  refcur in out sys_refcursor,
  in_table in varchar2,
  in_version_new in varchar2,
  in_version_old in varchar2
) is
  r_version_new_exist number := 0; 
  r_version_old_exist number := 0; 
begin
  begin
    select 1 into r_version_new_exist from d_versions where s_number = in_version_new and is_enabled = 1;
    exception when no_data_found then raise_application_error(-20100, 'Version "new" does not exist. Trace: '||dbms_utility.format_error_backtrace||dbms_utility.format_error_stack);
    return;
  end;
  
  begin
    select 1 into r_version_old_exist from d_versions where s_number = in_version_old and is_enabled = 1;
    exception when no_data_found then raise_application_error(-20101, 'Version "old" does not exist. Trace: '||dbms_utility.format_error_backtrace||dbms_utility.format_error_stack);
    return;
  end;
  
  if (r_version_new_exist = 0 or r_version_old_exist = 0) then
    raise_application_error(-20102, 'One of the versions does not exist. Trace: '||dbms_utility.format_error_backtrace||dbms_utility.format_error_stack);
  end if;

  open refcur for
    'with
    new_uniq as (select * from MV'|| in_version_new ||'_'|| in_table ||' minus select * from MV'|| in_version_old ||'_'|| in_table ||'),
    old_uniq as (select * from MV'|| in_version_old ||'_'|| in_table ||' minus select * from MV'|| in_version_new ||'_'|| in_table ||'),

    new_only as (select id from new_uniq minus select id from old_uniq),
    old_only as (select id from old_uniq minus select id from new_uniq),
    both_only as (select id from new_uniq intersect select id from old_uniq)

    -- inserted records
    select nu.*, ''i'' as oper# from new_uniq nu where exists(select 1 from new_only where id = nu.id)

    union all
    -- deleted records
    select ou.*, ''d'' as oper# from old_uniq ou where exists(select 1 from old_only where id = ou.id)

    union all
    -- updated records
    select nu.*, ''u'' as oper# from new_uniq nu where exists(select 1 from both_only where id = nu.id)';
    
    exception when others then raise_application_error(-20200, dbms_utility.format_error_backtrace||dbms_utility.format_error_stack);
end;
